import functools
import itertools
import json
import os
import os.path
import re

from typing import Generator, List, Iterator, Pattern, Any, Tuple, Dict, Match

import click

from .. import match as mat
from .. import config


def walk(dir: str, recursive: bool=False, match_directories: bool=False) -> Generator[os.DirEntry, None, None]:
    for entry in os.scandir(dir):
        is_dir = entry.is_dir()
        if recursive and is_dir:
            yield from walk(entry.path, recursive, match_directories)
        elif entry.is_file() or (match_directories and is_dir):
            yield entry
        else:
            continue


def apply_regex(
        walk_function: Iterator[Any],
        regex_pattern: Pattern,
        match_full_path: bool,
) -> Tuple[List[Dict], bool]:
    matches = []
    has_invalid_matches = False
    for f in walk_function:
        subject = f.path if match_full_path else f.name
        try:
            match = mat.apply(subject, regex_pattern)
        except ValueError:
            click.secho('{} did not match'.format(subject), fg='red', err=True)
            has_invalid_matches = True
        else:
            match['_path'] = f.path
            match['_name'] = f.name
            match['_abspath'] = os.path.abspath(f.path)
            match['_realpath'] = os.path.realpath(f.path)
            match['_relpath'] = os.path.relpath(f.path)
            matches.append(match)

    return matches, has_invalid_matches


@click.command(context_settings={'help_option_names': ('-h', '--help', '-?')})
@click.option('-d', '--directory', multiple=True, default=['./'])
@click.option('-m', '--match-directories', is_flag=True, default=False,
              help='search directory name also for matches')
@click.option('-r', '--recursive', is_flag=True, default=False,
              help='search directories recursively')
@click.option('--abort-on-no-match', is_flag=True, default=False,
              help='exit if a name does not match the INPUT_PATTERN')
@click.option('--match-full-path', is_flag=True, default=False,
              help='match on full path instead of only filename')
@click.option('-p', '--use-predefined-pattern', is_flag=True, default=False,
              help='use predefinged reges pattern')
@click.option('-v', '--verbose', is_flag=True, default=False,
              help='explain what is being done')
@click.version_option()
@click.argument('regex-pattern')
def pmatch(
    regex_pattern: str,
    directory: List[str],
    match_directories: bool,
    recursive: bool,
    abort_on_no_match: bool,
    match_full_path: bool,
    use_predefined_pattern: bool,
    verbose: bool,
) -> None:

    if use_predefined_pattern:
        regex_pattern, _ = config.load_template(regex_pattern, 'pmatch')

    try:
        regex_pattern_c = re.compile(regex_pattern)
    except re.error as e:
        if verbose:
            click.secho('Regular expression could not be compiled!', fg='red', err=True)
        click.secho(e.msg, fg='red', err=True)  # type: ignore
        raise SystemExit

    # Monkey patch walk function to include commandline options
    walk_function = functools.partial(walk, recursive=recursive, match_directories=match_directories)
    new_walk = itertools.chain.from_iterable(map(walk_function, directory))

    matches, has_invalid_matches = apply_regex(new_walk, regex_pattern_c, match_full_path)

    if has_invalid_matches and abort_on_no_match:
        raise SystemExit

    click.echo(json.dumps(matches))


if __name__ == '__main__':
    pmatch()
