import functools
import json
import multiprocessing
import os
import shlex
import subprocess

from typing import (
    Union,
    List,
    Dict,
    Callable,
)

import click

from .. import template
from .. import config
from .. import utils


def run_command(
    data: Dict[str, str],
    render: Callable[..., str],
    run: Callable[..., None],
    shell: bool,
) -> None:
    c = render(**data)
    if not shell:
        prepared_command = shlex.split(c)  # type: Union[str, List[str]]
    else:
        prepared_command = c

    run(prepared_command)


@click.command(context_settings={'help_option_names': ('-h', '--help', '-?')})
@click.option('--read-from', default='-', type=click.File('r'))
@click.option('-s', '--shell', is_flag=True, default=False,
              help='run command in a shell')
@click.option('-t', '--template-engine', default='python', type=click.Choice(['python', 'jinja2']),
              help='select the template engine to use')
@click.option('-v', '--verbose', is_flag=True, default=False,
              help='explain what is being done')
@click.option('-P', '--max-procs', default=1, type=click.IntRange(min=0, clamp=True))
@click.option('-p', '--use-predefined-template', is_flag=True, default=False,
              help='use predefined template')
@click.option('--redirect-stdout', is_flag=True, default=False,
              help='Redirect stdout to /dev/null')
@click.option('--redirect-stderr', is_flag=True, default=False,
              help='Redirect stderr to /dev/null')
@click.version_option()
@click.argument('command')
def targs(
    command: str,
    read_from: click.File,
    shell: bool,
    template_engine: str,
    max_procs: int,
    redirect_stdout: bool,
    redirect_stderr: bool,
    use_predefined_template: bool,
    verbose: bool,
) -> None:

    try:
        data = json.load(read_from)  # type: ignore
    except json.JSONDecodeError:
        click.secho('JSON not valid!', fg='red', err=True)
        raise SystemExit

    if use_predefined_template:
        command, template_engine = config.load_template(command, 'targs')

    try:
        render = template.get_render_function(command, engine=template_engine)
    except RuntimeError:
        click.secho('jinja2 not installed!', fg='red', err=True)
        raise SystemExit

    max_procs = utils.max_procs() if max_procs == 0 else max_procs

    stdout = subprocess.DEVNULL if redirect_stdout else None
    stderr = subprocess.DEVNULL if redirect_stderr else None
    patched_subprocess = functools.partial(
        subprocess.run,
        shell=shell,
        stdout=stdout,
        stderr=stderr,
    )
    patched_run_command = functools.partial(
        run_command,
        render=render,
        run=patched_subprocess,
        shell=shell,
    )
    if max_procs == 1:
        for r in data:
            patched_run_command(r)
    else:
        with multiprocessing.Pool(max_procs) as p:
            p.map(patched_run_command, data)


if __name__ == '__main__':
    targs()
