import json
import os
import os.path

import click

from .. import template
from .. import config


def link_file(old_filename: str, new_filename: str, symbolic: bool) -> None:
    dirs = os.path.dirname(new_filename)
    if dirs:
        os.makedirs(dirs, exist_ok=True)

    if symbolic:
        os.symlink(old_filename, new_filename)
    else:
        os.link(old_filename, new_filename)


@click.command(context_settings={'help_option_names': ('-h', '--help', '-?')})
@click.option('-i', '--interactive', is_flag=True, default=False,
              help='prompt wether to remove destinations')
@click.option('-f', '--force', is_flag=True, default=False,
              help='remove existing destination files')
@click.option('-s', '--symbolic', is_flag=True, default=False,
              help='make symbolic links instead of hard links')
@click.option('-t', '--template-engine', default='python', type=click.Choice(['python', 'jinja2']),
              help='select the template engine to use')
@click.option('-p', '--use-predefined-template', is_flag=True, default=False,
              help='use predefined template')
@click.option('--read-from', default='-', type=click.File('r'))
@click.option('--abort-on-path-exist', is_flag=True, default=False,
              help='exit if a renamed path already exists')
@click.option('-n', '--dry-run', is_flag=True, default=False,
              help='perform a trial run with no changes made')
@click.option('-v', '--verbose', is_flag=True, default=False,
              help='explain what is being done')
@click.version_option()
@click.argument('output-pattern')
def tln(
    output_pattern: str,
    interactive: bool,
    force: bool,
    symbolic: bool,
    template_engine: str,
    use_predefined_template: bool,
    read_from: click.File,
    abort_on_path_exist: bool,
    dry_run, bool,
    verbose: bool,
) -> None:
    pass

    try:
        data = json.load(read_from)  # type: ignore
    except json.JSONDecodeError:
        click.secho('JSON not valid!', fg='red', err=True)
        raise SystemExit

    if use_predefined_template:
        output_pattern, template_engine = config.load_template(output_pattern, 'tmv')

    try:
        render = template.get_render_function(output_pattern, engine=template_engine)
    except RuntimeError:
        click.secho('jinja2 not installed!', fg='red', err=True)
        raise SystemExit

    new_filenames = dict((entry['_abspath'], render(**entry)) for entry in data)

    has_existing = False
    non_existing = {}
    for old, new in new_filenames.items():
        if os.path.exists(new):
            click.secho('{} does exist!'.format(new), fg='red', err=True)
            has_existing = True
        else:
            non_existing[old] = new

    if has_existing and abort_on_path_exist:
        raise SystemExit

    if force:
        files_to_move = new_filenames.items()
    else:
        files_to_move = non_existing.items()

    for old, new in files_to_move:
        if verbose or interactive:
            click.echo('{} -> {}'.format(old, new))

        if not (interactive and not click.confirm('Is the match correct?')):
            if not dry_run:
                link_file(old, new, symbolic)
